import { COUNTER_LENGTH } from './constants'
import { tweakeyEncode, skinnyEncrypt } from './skinny-128-384-plus'

/**
 * Parse message into blocks.
 * @param message The message to parse.
 * @param blockLength The block length.
 * @returns An array of blocks.
 */
function parse (message: number[], blockLength: number): number[][] {
  // Keep track of position in message currently parsed into blocks.
  let cursor = 0

  // Slice message into blocks.
  let ret: number[][] = []
  while (message.length - cursor >= blockLength) {
    ret.push(...[message.slice(cursor, cursor + blockLength)])
    cursor = cursor + blockLength
  }

  // Append any remaining blocks regardless of block length. These will be padded later.
  if (message.length - cursor > 0) {
    ret.push(...[message.slice(cursor)])
  }

  // If no message, return a single block.
  if (message.length === 0) {
    ret = [[]]
  }

  // Insert empty array at position 0.
  ret.splice(0, 0, [])
  return ret
}

/**
 * Pads the byte length of message to padLength. The final byte (when padded) contains the original message length.
 * @param message
 * @param padLength
 * @returns A complete 16 byte block.
 */
function pad (message: number[], padLength: number): number[] {
  // If there is no message, return a fully padded block.
  if (message.length === 0) {
    return Array(16).fill(0)
  }

  // Return a copy of the message if no padding is required.
  if (message.length === padLength) {
    return [...message]
  }

  // Pad a copy of the message to padLength.
  const ret = [...message]
  const requiredPadding = padLength - message.length - 1
  ret.push(...Array(requiredPadding).fill(0))

  // Set the final byte of the padded blocked to the length of the original message.
  ret[padLength - 1] = message.length

  return ret
}

/**
 * Generate the key stream from the internal state by multiplying the state S and the constant matrix G.
 * @param state
 * @returns The key stream.
 */
function g (state: number[]): number[] {
  return state.map(x => {
    return (x >> 1) ^ (x & 0x80) ^ ((x & 0x01) << 7)
  })
}

/**
 * The state update function. Pads an M block.
 * @param state The internal state, S.
 * @param mBlock An M block.
 * @returns [S', C] where S' = M ⊕ S and C = M ⊕ G(S)
 */
function rho (state: number[], mBlock: number[]): [number[], number[]] {
  // G(S)
  const gOfS = g(state)

  // C = M ⊕ G(S)
  const c = [...Array(16).keys()].map(i => mBlock[i] ^ gOfS[i])

  // S' = M ⊕ S
  const nextState = [...Array(16).keys()].map(i => state[i] ^ mBlock[i])

  return [nextState, c]
}

/**
 * Increments the 56 bit LFSR-based counter.
 * @param counter The old counter.
 * @returns An incremented counter.
 */
function increaseCounter (counter: number[]): number[] {
  const fb0 = counter[6] >> 7

  counter[6] = (counter[6] << 1) | (counter[5] >> 7)
  counter[5] = (counter[5] << 1) | (counter[4] >> 7)
  counter[4] = (counter[4] << 1) | (counter[3] >> 7)
  counter[3] = (counter[3] << 1) | (counter[2] >> 7)
  counter[2] = (counter[2] << 1) | (counter[1] >> 7)
  counter[1] = (counter[1] << 1) | (counter[0] >> 7)

  if (fb0 === 1) {
    counter[0] = (counter[0] << 1) ^ 0x95
  } else {
    counter[0] = (counter[0] << 1)
  }

  return counter
}

/**
 * Returns a reset counter.
 * @returns A reset counter.
 */
function resetCounter (): number[] {
  const counter = Array(COUNTER_LENGTH).fill(0)
  counter[0] = 1
  return counter
}

/**
 * Returns a zeroed buffer.
 * @param bufferLength The length of the buffer to return.
 * @returns A zeroed buffer.
 */
function zeroedBuffer (bufferLength: number): number[] {
  return Array(bufferLength).fill(0)
}

/**
 * Calculate the domain separation.
 * @param combinedData The parsed and concatenated message and associated data,
 * @param parsedMessageLength The length of the parsed message.
 * @param parsedAssociatedDataLength The length of the parsed associated data.
 */
function calculateDomainSeparation (combinedData: number[][], parsedMessageLength: number, parsedAssociatedDataLength: number): number {
  let domainSeparation = 16

  if (combinedData[parsedAssociatedDataLength].length < 16) {
    domainSeparation = domainSeparation ^ 2
  }

  if (combinedData[parsedAssociatedDataLength + parsedMessageLength].length < 16) {
    domainSeparation = domainSeparation ^ 1
  }

  if (parsedAssociatedDataLength % 2 === 0) {
    domainSeparation = domainSeparation ^ 8
  }

  if (parsedMessageLength % 2 === 0) {
    domainSeparation = domainSeparation ^ 4
  }

  return domainSeparation
}

/**
 * Encrypt a message using the Romulus-M cryptography specification.
 * @param message The message to encrypt.
 * @param associatedData The associated data to encrypt.
 * @param nonce A 128 bit nonce.
 * @param key A 128 bit encryption key.
 * @returns The ciphertext.
 */
export function cryptoAeadEncrypt (message: number[], associatedData: number[], nonce: number[], key: number[]): number[] {
  let counter = resetCounter()

  // Carve message and associated data into blocks.
  const parsedMessage = parse(message, 16)
  const parsedMessageLength = parsedMessage.length - 1

  const parsedAssociatedData = parse(associatedData, 16)
  const parsedAssociatedDataLength = parsedAssociatedData.length - 1

  // Concatenate the parsed message and the associated data, excluding each array's first element.
  const combinedData = parsedAssociatedData.slice(1).concat(parsedMessage.slice(1))

  // Insert empty array at position 0.
  combinedData.splice(0, 0, [])

  // Calculate domain separation for final encryption stage.
  const domainSeparation = calculateDomainSeparation(combinedData, parsedMessageLength, parsedAssociatedDataLength)

  // Pad combined data.
  combinedData[parsedAssociatedDataLength] = pad(combinedData[parsedAssociatedDataLength], 16)
  combinedData[parsedAssociatedDataLength + parsedMessageLength] = pad(combinedData[parsedAssociatedDataLength + parsedMessageLength], 16)

  // Do the encryption.
  // See https://romulusae.github.io/romulus/docs/Romulusv1.3.pdf for more information.
  let state = zeroedBuffer(16)
  let c = zeroedBuffer(16)
  let x = 8
  for (let i = 1; i < Math.floor((parsedAssociatedDataLength + parsedMessageLength) / 2) + 1; i++) {
    [state, c] = rho(state, combinedData[2 * i - 1])
    counter = increaseCounter(counter)
    if (i === Math.floor(parsedAssociatedDataLength / 2) + 1) {
      x = x ^ 4
    }
    state = skinnyEncrypt(state, tweakeyEncode(counter, x, combinedData[2 * i], key))
    counter = increaseCounter(counter)
  }

  if (parsedAssociatedDataLength % 2 === parsedMessageLength % 2) {
    [state, c] = rho(state, zeroedBuffer(16))
  } else {
    [state, c] = rho(state, combinedData[parsedAssociatedDataLength + parsedMessageLength])
    counter = increaseCounter(counter)
  }

  [state, c] = rho(skinnyEncrypt(state, tweakeyEncode(counter, domainSeparation, nonce, key)), zeroedBuffer(16))
  if (message.length === 0) {
    return c
  }

  state = [...c]

  const ciphertext = []
  const originalLastBlockLength = parsedMessage[parsedMessageLength].length
  parsedMessage[parsedMessageLength] = pad(parsedMessage[parsedMessageLength], 16)

  counter = resetCounter()
  for (let i = 1; i < parsedMessageLength + 1; i++) {
    state = skinnyEncrypt(state, tweakeyEncode(counter, 4, nonce, key))
    let x
    [state, x] = rho(state, parsedMessage[i])
    counter = increaseCounter(counter)
    if (i < parsedMessageLength) {
      ciphertext.push(...x)
    } else {
      ciphertext.push(...x.slice(0, originalLastBlockLength))
    }
  }

  ciphertext.push(...c)

  return ciphertext
}
